// display

// the following lines is supposed(optional) to replace a single line in sketch, the #include"Esp8266MyWiFiUnified.h".
// For the following macros copied to main for easy modify. So, copy these to there and uncomment it could simply done.
#ifndef _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
#define _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
#include <ESP8266WiFi.h>

////////////////// paste the macros here (which is between IoT BEGIN/END .h) //////////////////
// ************************** IoT BEGIN .h ******************************
// ========================== Options and Defs/modify if need ==============================
// -------------------------- for AP or STA ------------------------------
// 1 for ap, 0 for sta.
#define AP_MODE 0
#define STA_MODE (!AP_MODE)

// -------------------------- for ap/sta credential ------------------------------ comment it and put it in .cpp
#if 1
const char *const WIFI_CRED[][2]={
    {"ESP_TESTAP", "87654321"},
    {"Office", "123456"},
    {"Xperia X Performance_bc10", "e848acbcfd98"},
};
#endif // if 1

// -------------------------- for part of hostname ------------------------------
#define LSERVICE_1 "display"
#define LPSERVICE_1 80

// -------------------------- for default connection names ------------------------------
// i.e., http://hostname.local/
#define HOSTNAME_BASE "esp8266-iot" // for local used.
#define HOSTNAME ((String(HOSTNAME_BASE)+"-"+LSERVICE_1).c_str()) // for local used.
#define REMOTE_HOSTNAME "esp8266-iot-internet" // for remote use.

#define UPDFILE "a.bin" // the remote webpage for updating to local is http://REMOTE_HOSTNAME.local/upd

// -------------------------- other settings ------------------------------
#define HTTPC_GET_DELAY delay(25) // do the delay after an HTTP GET.
#define HTTPC_GET_PROC_DELAY delay(5) // do some delay during an HTTP GET course.
#define HTTPC_CONN_TIMEOUT 3000 // the timeout during a GET, 3000ms.

// all the TMO values is restricted to no more than 35 minutes; see the Timeout() for detail. however,
// cascading could overcome.
#define TMO_IOT_LOOP_PERIOD 75 // IoT polling at each 75ms
#define TMO_AP_C_CONN_PERIOD 90000 // each 90 seconds to show connections by AP
#define TMO_STA_CONN_HEALTH_PERIOD 20000 // each 20 seconds to check the connection health by STA(while ping is alive)
#define TMO_WEB_UPDATE_PERIOD (1000*60*30) // each 30 minutes to find internet update; if cascade twice(no, use count) will become 15hr
#define TMO_MDNS_Q_PERIOD 120000 // each 120 seconds to query services
#define TMO_PING_PERIOD 5000 // each 5 seconds to ping AP when ping failed

#define STA_RECOVER_ALIVE 0 // set 0 if STA connection broken and do not reconnect. 1 to reconnect.

// [mDns] use either of one
#define USE_MDNS_STANDALONE // no auto hostname
//#define USE_MDNS_FULL // auto hostname and query

// [STAtion] use either of one
//#define USE_STA_NATIVE // single AP
#define USE_STA_WIFI_MULTI // multiple APs(o)

// [HTTPClient] use either of one
//#define USE_HTTPC_HTTP_CLIENT
//#define USE_HTTPC_WIFI_CLIENT
#define USE_HTTPC_WIFIHTTP_HYBRID // (o2)
//#define USE_HTTPC_WIFIHTTP_HYBRID_NONBLOCKING // (o1)

// -------------------------- for mdns ------------------------------
typedef struct smDnsRecord{
    const char *service;
    const unsigned port;
} smDnsRecord;

// -------------------------- for mdns services ------------------------------ comment it and put it in .cpp
#if 1
#if defined(USE_MDNS_FULL)
const smDnsRecord mdns_ss[]={ // the services this device provides; note, at least the LSERVICE_1 has already provided.
    {"display", 80},
    {"iot", 80},
    {"https", 443},
    {"test", 81},
    {"motor", 80},
};
#endif // USE_MDNS_FULL

const smDnsRecord mdns_sskn[]={ // the known services this device needs from other devices.
    {"display", 80},
    {"iot", 80},
    {"https", 443},
    {"test", 81},
    {"motor", 80},
};
#endif // if 1


// ************************** IoT END .h ******************************
////////////////// paste the macros here (which is between IoT BEGIN/END .h) //////////////////

#include"Esp8266MyWiFiUnified.h"
#endif // _ESP8266_MY_WIFI_UNIFIED_DISABLEH_


#include <U8g2lib.h>
#include"Esp8266MultiPwms.h"
#include"Esp8266Vars2File.h"

// note: http command: stop timer, before web update.
// currently uses white version board.
// known issues: 1) IRAM_ATTR ran out. 2) ping causes wdt-reset.


#define VERSION_202105180003 "[[202105180003][AP/STA][ESP_TESTAP:87654321][esp8266-iot-display:80][esp8266-iot-internet][upd:a.bin]"\
            "[MDNS:FIXED][non WeMos TTGO][115200:160:26:4M:2MFS:1MSK:80M]]" // so the version number is the 3rd byte, length 12.


String web_link, ap_mode_speed_str;
unsigned stop_the_timer=0;
unsigned poweroff=0;
String netstr_ssid, netstr_localip, netstr_remoteip, netstr_localhostname;
IRAM_ATTR extern void delayMicroseconds(unsigned);
IRAM_ATTR extern void analogRead(unsigned);
unsigned copy_sta_do_the_connect; // default is 0. for STA connection. since order in main() is of concern, so copy.

// ------------------------------------ timer settings
#define DRAW_INT_GPIO           12
#define DRAW_TRG_GPIO           13
#define DRAW_PERIOD             80000 // 80ms. less than 80ms will wdt-reset when ap mode.
#define BTN_PERIOD              20000 // 20ms. less than 20ms will wdt-reset when ap mode.


// ------------------------ mux
// mux s2=gnd, s1=gpio13, s0=gpio12, use a0, a1, a2, a3 for y, x, button, battery.

//#define MUX_S0_GPIO             16 // TTGO
#define MUX_S0_GPIO             2 // non TTGO
#define MUX_S1_GPIO             3

//#define BUTTON_GPIO             5 // TTGO
#define BUTTON_GPIO             14 // non TTGO

#define VOLTAGE_REAL            14.5f
#define VOLTAGE_TRAN            3.248f
#define VOFFSET                 17
#define BVOLTAGE(x)             (((x)>VOFFSET? (x)-17: 0)*(3.3f/VOLTAGE_TRAN*VOLTAGE_REAL/1024.f))


// ------------------------------------ u8g2 config
// TTGO
//#define OLED_SDA  2
//#define OLED_SCL 14
//#define OLED_RST  4

// non TTGO
#define OLED_SDA  4
#define OLED_SCL  5
#define OLED_RST 16

U8G2_SSD1306_128X32_UNIVISION_F_SW_I2C u8g2(U8G2_R0, OLED_SCL, OLED_SDA, OLED_RST);

void setup_u8g2(){
    u8g2.begin();
}


// ------------------------------------ timer processing for both joystick and u8g2 drawing.

// to store draw functions and for traversal.
// each linkedlist draws a page(of course the present page), each node draws parts of the page. order is of concerned.
typedef struct sFooLinkedList{
    void (*fp)();
    sFooLinkedList *next;
    sFooLinkedList(): fp(0), next(0){;};
    ~sFooLinkedList(){delete this->next;};
} sFooLinkedList;

cHwTimer timer_draw_manip;                  // the timer obj. be aware that hwtimer1 must bring up first.
IRAM_ATTR unsigned timer_trigger=0;         // the variable for toggling.
IRAM_ATTR void isr_draw_manip_timer(){digitalWrite(DRAW_TRG_GPIO, timer_trigger^=1);} // the timer isr.

IRAM_ATTR sFooLinkedList *draw_page_1=0;    // the page presently.
IRAM_ATTR sFooLinkedList *draw_page_2=0;    // the incoming page. in order for page sliding animation.
/**IRAM_ATTR **/int dpox=0; // the draw page offset, x. in order for page sliding animation. use IRAM_ATTR causes u8g2.drawBox fault.
/**IRAM_ATTR **/int dpoy=0; // the draw page offset, y. in order for page sliding animation. use IRAM_ATTR causes u8g2.drawBox fault.
IRAM_ATTR int slide_points_x[]={5, 16, 28, 48, 20, 6, 5};   // for slide x animation.
IRAM_ATTR int slide_points_y[]={2, 3, 6, 15, 3, 2, 1};      // for slide y animation.
IRAM_ATTR int slide_idx;    // keeps the index while animating.
#define SLIDE_DIR_TIPIDLE   0
#define SLIDE_DIR_TIPLEFT   1
#define SLIDE_DIR_TIPRIGHT  -1
#define SLIDE_DIR_TIPDOWN   -2
#define SLIDE_DIR_TIPUP     2
IRAM_ATTR int slide_dir;    // keeps the direction while animating. (-2, down), (-1, right), (1, left), (2, up).
IRAM_ATTR int slide_2_pos;  // it mainly presenting the zero position of the incoming page. nonzero means still animating.
IRAM_ATTR unsigned DrawSubmit=0;            // set 1 to start drawing/refresh.
IRAM_ATTR unsigned AutoDrawSubmit=0;        // set 1 to start auto drawing/refresh.

IRAM_ATTR unsigned joystick_vol_old;        // the previously battery voltage adc value.
IRAM_ATTR unsigned joystick[4];             // 4 channels ADC values. (0, y), (1, x), (2, button), (3, battery).
IRAM_ATTR unsigned joystick_tri[4];         // 4 channels ADC values of changes. if changes that channel has triggered.
IRAM_ATTR unsigned joystick_alive=0;        // any channel has triggered is alive. all channels unchanged is not alive.
unsigned joystick_x_old, joystick_y_old, joystick_b_old; // x, y, batt. used for probing being triggered(changed) or not.
#define JS_DIR_IDLE         0
#define JS_DIR_LEFT         1
#define JS_DIR_RIGHT        2
#define JS_DIR_DOWN         4
#define JS_DIR_UP           8
#define JS_DIR_PUSH         16
unsigned joystick_dir; // idle=0, left=1, right=2, down=4, up=8, push=16.
unsigned joystick_action; // (0, idle); (1, under acting).

void DrawSetup(){ // for timer isr and gpio isr. initialize once.
    pinMode(DRAW_TRG_GPIO, OUTPUT); // this gpio is to trigger intr gpio.
    digitalWrite(DRAW_TRG_GPIO, LOW);
    attachInterrupt(digitalPinToInterrupt(DRAW_INT_GPIO), isr_draw_manip, CHANGE); // setup gpio interrupt.
}

void DrawConfig(class sFooLinkedList *&draw_link, void (*fptr)()){ // insert function pointers to which draws the page buffer.
    if (!draw_link){
        draw_link=new sFooLinkedList;
        draw_link->fp=fptr;
        return;
    }
    sFooLinkedList *a_link=draw_link;
    while (a_link->next) a_link=a_link->next;
    a_link->next=new sFooLinkedList;
    a_link->next->fp=fptr;
}

void DrawEnd(class sFooLinkedList *&draw_link){ // must called if do not draw anymore.
    // note that violations(in isr) do exist. advanced treatment if needed. e.g., do it in isr or use gate.
    unsigned i=DrawSubmit, j=AutoDrawSubmit;
    AutoDrawSubmit=0;
    DrawSubmit=2;
    delete draw_link;
    draw_link=0;
    DrawSubmit=i;
    AutoDrawSubmit=j;
}

IRAM_ATTR void DrawPageUpdateOffset_1(){ // handling while animating for present page.
    int i;
    dpox=dpoy=0;
    switch (slide_dir){
        case SLIDE_DIR_TIPDOWN:
            for (i=0; i<=slide_idx; i++) dpoy+=slide_points_y[i];
            dpoy=0-dpoy;
            break;
        case SLIDE_DIR_TIPRIGHT:
            for (i=0; i<=slide_idx; i++) dpox+=slide_points_x[i];
            dpox=0-dpox;
            break;
        case SLIDE_DIR_TIPIDLE: break; // null
        case SLIDE_DIR_TIPLEFT:
            for (i=0; i<=slide_idx; i++) dpox+=slide_points_x[i];
            break;
        case SLIDE_DIR_TIPUP:
            for (i=0; i<=slide_idx; i++) dpoy+=slide_points_y[i];
            break;
    }
}

IRAM_ATTR void DrawPageUpdateOffset_2(){ // handling while animating for incoming page.
    switch (slide_dir){
        case SLIDE_DIR_TIPDOWN:
            slide_2_pos-=slide_points_y[slide_idx];
            dpoy=slide_2_pos;
            break;
        case SLIDE_DIR_TIPRIGHT:
            slide_2_pos-=slide_points_x[slide_idx];
            dpox=slide_2_pos;
            break;
        case SLIDE_DIR_TIPIDLE: break; // null
        case SLIDE_DIR_TIPLEFT:
            slide_2_pos+=slide_points_x[slide_idx];
            dpox=slide_2_pos;
            break;
        case SLIDE_DIR_TIPUP:
            slide_2_pos+=slide_points_y[slide_idx];
            dpoy=slide_2_pos;
            break;
    }
    slide_idx++;
}

IRAM_ATTR void demandcall_draw_manip(){
    if (DrawSubmit==1){
        DrawSubmit=0;
        u8g2.clearBuffer();
        sFooLinkedList *a_link=draw_page_1;
        if (slide_2_pos) DrawPageUpdateOffset_1();
        while (a_link){
            if (a_link->fp) (a_link->fp)();
            a_link=a_link->next;
        }
        if (slide_2_pos){
            a_link=draw_page_2;
            DrawPageUpdateOffset_2();
            while (a_link){
                if (a_link->fp) (a_link->fp)();
                a_link=a_link->next;
            }
            if (!slide_2_pos) DrawSubmit=2; // in order to hold the refresh
        }
        u8g2.sendBuffer();
    }
    if (AutoDrawSubmit && DrawSubmit!=2) DrawSubmit=1;
}

IRAM_ATTR void isr_draw_manip(){

    // ------ for drawing
    IRAM_ATTR static unsigned t=0;
    t+=BTN_PERIOD;
    if (t>=DRAW_PERIOD){
        t=0;
        demandcall_draw_manip();
    }

    // ------ for button sensing
    // joystick sampling. CD74HC4051E Mux.
    // a0, a1, a2, a3 for y, x, button, battery.

    IRAM_ATTR static unsigned pn=0;
    digitalWrite(MUX_S0_GPIO, pn&0x1);
    digitalWrite(MUX_S1_GPIO, (pn>>1)&0x1);
    delayMicroseconds(1);
    joystick[pn]=analogRead(A0);

    if (joystick[0]){ // the joystick is now working
        switch (pn){
            case 0: // y
                if (joystick[0]<300){
                    joystick_tri[0]+=((joystick_tri[0]&0x1)? 2: 1); // odd if dir=-y.
                    joystick_alive=1;
                }
                else if (joystick[0]>700){
                    joystick_tri[0]+=((joystick_tri[0]&0x1)? 1: 2); // even if dir=+y.
                    joystick_alive=1;
                }
                break;
            case 1: // x
                if (joystick[1]<300){
                    joystick_tri[1]+=((joystick_tri[1]&0x1)? 2: 1); // odd if dir=-x.
                    joystick_alive=1;
                }
                else if (joystick[1]>700){
                    joystick_tri[1]+=((joystick_tri[1]&0x1)? 1: 2); // even if dir=+x.
                    joystick_alive=1;
                }
                break;
            case 2: // button
                if (joystick[2]>300){ // caught the make code
                    joystick_tri[2]++;
                    joystick_alive=1;
                }
                break;
            case 3: // battery voltage, low detection.
                // around the battery voltage detection of 3.45V, the LDO will output 2.63V and can not boot up.
                // so the threshold sould be 3.7V(LDO 2.9V).
                extern const char *bat_status;
                joystick[3]=joystick_vol_old*0.8f+joystick[3]*0.2f;
                joystick_vol_old=joystick[3];
                if (BVOLTAGE(joystick[3])<3.7f){ // battery low
                    if (joystick_tri[3]<1000) joystick_tri[3]++;
                    else if (joystick_tri[3]==1000){
                        bat_status="l"; // 4
                        joystick_tri[3]++;
                    }
                }
                else if (joystick_tri[3]){
                    if (joystick_tri[3]==1) bat_status="-"; // 0
                    joystick_tri[3]--;
                }
                break;
        }
    }
    if (++pn>3) pn=0;
}


// ------------------------------------ u8g2 drawing
#define U8G2_X          128
#define U8G2_Y          32
IRAM_ATTR unsigned countdown=0;
int mapx_idx, mapy_idx;
unsigned locked_lr;
IRAM_ATTR int jtk_pos_x, jtk_pos_y;
/*IRAM_ATTR */String input_str; // not for IRAM_ATTR
const char alphabet[][26]={
    "abcdefghijklm ~!@#$%^&_()",
    "nopqrstuvwxyz`+-*/=\\'?\"{}",
    "LUUCBS1234567890<>,;.:|[]"
}; // LUUCBS, line-feed, upper-once, upper-fixed, clear, backspace, save.

// ==== u8g2.setFont(u8g2_font_5x7_tr); === occupy 5x7 pixels === need not preserve empty pixel ======= 128x32 ======
// in a row, positioning start from 1, end to 121, 25 characters
// in a col, positioning start from 2(8), end to 23(29), 4 characters
// first col, (1, 8), (1, 15), (1, 22), (1, 29)

// ==== u8g2.setFont(u8g2_font_8x13B_mf); === occupy 8x12 pixels === need not preserve empty pixel ======= 128x32 ======
// in a row, positioning start from 0, end to 120, 16 characters
// in a col, positioning start from 1(13), end to 14(26), 2 characters
// first col, (0, 13), (0, 26)

// ==== u8g2.setFont(u8g2_font_unifont_t_chinese1); === 1 to 3 == 128x32 ======
// u8g2.enableUTF8Print(); first
// u8g2.setCursor(x, y);
// u8g2.print("");
// or u8g2.drawUTF8(x, y, "");

// ==== u8g2.setFont(u8g2_font_unifont_t_symbols); ===== 128x32 ======
// u8g2.enableUTF8Print(); first
// u8g2.drawUTF8(0, 28, "Snowman: ☃");

IRAM_ATTR void loop_u8g2_alphabet(){
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawStr(dpox+1,  dpoy+8, alphabet[0]);
    u8g2.drawStr(dpox+1, dpoy+15, alphabet[1]);
    u8g2.drawStr(dpox+1, dpoy+22, alphabet[2]);
    u8g2.drawBox(dpox+11, dpoy+16, 5, 7); // the LU"U"BS.
    u8g2.setColorIndex(0);
    u8g2.drawStr(dpox+11, dpoy+22, "U");
    u8g2.setColorIndex(1);
    /// could instead use this
    ///u8g2.setDrawColor(0);
    ///u8g2.drawStr(11, 22, "U");
    ///u8g2.setDrawColor(1);
    //u8g2.drawStr(103, 31, "|||||"); // u8g2.drawStr(103, 31, "nxt->"); // u8g2.drawStr(103, 31, "__|__");
}

/*IRAM_ATTR */void loop_u8g2_alphabet_change(){
    int a=jtk_pos_x*5+1;
    int b=(jtk_pos_y+1)*7+1;
    u8g2.drawBox(dpox+a, dpoy+b-7, 5, 7);
    u8g2.drawBox(dpox+1, dpoy+22, 5*input_str.length(), 7);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+a, dpoy+b, String(alphabet[jtk_pos_y][jtk_pos_x]).c_str());
    u8g2.drawStr(dpox+1, dpoy+29, input_str.c_str());
    u8g2.setDrawColor(1);
}

/*IRAM_ATTR */void loop_u8g2_welcome(){
    u8g2.setFont(u8g2_font_luBIS18_tr);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+4, dpoy+25, "welcome");
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
}

/*IRAM_ATTR */void loop_u8g2_poweroff(){
    u8g2.setFont(u8g2_font_luBIS18_tr);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+4, dpoy+25, "off now");
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
}

IRAM_ATTR void loop_u8g2_settings(){ // settings
    u8g2.setFont(u8g2_font_luBIS18_tr);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+4, dpoy+25, "SETTING");
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
}

/*IRAM_ATTR */void loop_u8g2_settings_poweronconnect(){ // settings->poweronconnect
    u8g2.setFont(u8g2_font_luBIS18_tr);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+4, dpoy+25, "PrOnCnn");
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
}

IRAM_ATTR void loop_u8g2_powerdown(){
}

IRAM_ATTR void loop_u8g2_status_051(){ // connection status; c/d connected/disconnected.
    u8g2.setFont(u8g2_font_5x7_tr);
#if STA_MODE
    if (!is_station_mode_established) u8g2.drawStr(103, 31, "d");
    else u8g2.drawStr(103, 31, "c");
#else // AP_MODE
    u8g2.drawStr(103, 31, "a"); // means AP mode
#endif // STA_MODE
}

IRAM_ATTR void loop_u8g2_status_052(){ // battery status.
    extern const char *bat_status;
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawStr(108, 31, bat_status);
}

IRAM_ATTR void loop_u8g2_status_053(){ // null
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawStr(113, 31, "-");
}

IRAM_ATTR void loop_u8g2_status_054(){ // null
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawStr(118, 31, "-");
}

IRAM_ATTR void loop_u8g2_status_055(){ // countdown to poweroff
    u8g2.setFont(u8g2_font_5x7_tr);
    if (countdown) u8g2.drawStr(123, 31, String(countdown).c_str());
    else u8g2.drawStr(123, 31, "a"); // means active
}

/*IRAM_ATTR */void loop_u8g2_on(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+35, dpoy+13, "NOW ON");
}

/*IRAM_ATTR */void loop_u8g2_off(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+35, dpoy+26, "NOW OFF");
}

IRAM_ATTR void loop_u8g2_1(){ // "U8g2" animation
    u8g2.setFont(u8g2_font_8x13B_mf);
    IRAM_ATTR static int x=-8, y=8;
    u8g2.drawStr(dpox+(x+=y), dpoy+13, "U8g2");
    if (x>=127-32) y=-8;
    else if (x<=0) y=8;
}

IRAM_ATTR void loop_u8g2_2(){ // show the speed value
    u8g2.setFont(u8g2_font_inr27_t_cyrillic);
    static String pr="";
    if (http_got_str!="" && pr!=http_got_str){
        pr=http_got_str;
    }

#if AP_MODE
    pr=ap_mode_speed_str;
#endif

    if (pr!="") u8g2.drawStr(dpox+15, dpoy+27, pr.c_str());
}

IRAM_ATTR void loop_u8g2_3(){ // output the joystick values
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+0, dpoy+26, (String(BVOLTAGE(joystick[3]))+" "+joystick_tri[1]+" "+joystick_tri[0]+" "+joystick_tri[2]).c_str());
}

IRAM_ATTR void loop_u8g2_page_1(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+0, dpoy+13, "page 1");
}

/*IRAM_ATTR */void loop_u8g2_page_2(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+0, dpoy+13, "page 2");
}

/*IRAM_ATTR */void loop_u8g2_page_3(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+0, dpoy+13, "page 3");
}

IRAM_ATTR void loop_u8g2_connect(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+24, dpoy+13, "connection");
    u8g2.setFont(u8g2_font_5x7_tr);
#if STA_MODE
    u8g2.drawStr(dpox+2, dpoy+22,\
        is_station_mode_established?\
        (String("-> connected: ")+netstr_ssid).c_str():\
        "-> disconnected."\
    );
#endif // STA_MODE
    u8g2.drawStr(dpox+0, dpoy+29, "tap to connect/disconnect");
}

IRAM_ATTR void loop_u8g2_network(){
    u8g2.setFont(u8g2_font_5x7_tr);
    #if STA_MODE
        u8g2.drawStr(dpox+1, dpoy+8,\
            is_station_mode_established?\
            (String("Local IP: ")+netstr_localip).c_str():\
            "Local IP:"\
        );
        u8g2.drawStr(dpox+1, dpoy+15,\
            is_station_mode_established?\
            (String("AP IP: ")+netstr_remoteip).c_str():\
            "AP IP:"\
        );
        u8g2.drawStr(dpox+1, dpoy+22,\
            is_station_mode_established?\
            (String("Hostname: ")+netstr_localhostname).c_str():\
            "Hostname:"\
        );
    #endif // STA_MODE
    u8g2.drawStr(dpox+1, dpoy+29, "_");
}

/*IRAM_ATTR */void loop_u8g2_network_cont(){
    u8g2.setFont(u8g2_font_5x7_tr);
    #if STA_MODE
        u8g2.drawStr(dpox-U8G2_X, dpoy+8,\
            is_station_mode_established?\
            (String("Local IP: ")+netstr_localip).c_str():\
            "Local IP:"\
        );
        u8g2.drawStr(dpox-U8G2_X, dpoy+15,\
            is_station_mode_established?\
            (String("AP IP: ")+netstr_remoteip).c_str():\
            "AP IP:"\
        );
        u8g2.drawStr(dpox-U8G2_X, dpoy+22,\
            is_station_mode_established?\
            (String("Hostname: ")+netstr_localhostname).c_str():\
            "Hostname:"\
        );
    #endif // STA_MODE
    u8g2.drawStr(dpox-U8G2_X, dpoy+29, "_");
}

void page_u8g2_subscript(){
    DrawConfig(draw_page_2, loop_u8g2_status_051);
    DrawConfig(draw_page_2, loop_u8g2_status_052);
    DrawConfig(draw_page_2, loop_u8g2_status_053);
    DrawConfig(draw_page_2, loop_u8g2_status_054);
    DrawConfig(draw_page_2, loop_u8g2_status_055);
}

void page_u8g2_mainmenu(){
    DrawConfig(draw_page_2, loop_u8g2_1);
    DrawConfig(draw_page_2, loop_u8g2_3);
    page_u8g2_subscript();
}

void page_u8g2_page1(){
    DrawConfig(draw_page_2, loop_u8g2_page_1);
    page_u8g2_subscript();
}

void page_u8g2_page2(){
    DrawConfig(draw_page_2, loop_u8g2_page_2);
    page_u8g2_subscript();
}

void page_u8g2_page3(){
    DrawConfig(draw_page_2, loop_u8g2_page_3);
    page_u8g2_subscript();
}

void page_u8g2_speed(){
    DrawConfig(draw_page_2, loop_u8g2_2);
    page_u8g2_subscript();
}

void page_u8g2_input(){
    DrawConfig(draw_page_2, loop_u8g2_alphabet);
}

void page_u8g2_connect(){
    DrawConfig(draw_page_2, loop_u8g2_connect);
}

void page_u8g2_network(){
    DrawConfig(draw_page_2, loop_u8g2_network);
}

void page_u8g2_network_cont(){
    DrawConfig(draw_page_2, loop_u8g2_network_cont);
}

void page_u8g2_settings(){
    DrawConfig(draw_page_2, loop_u8g2_settings);
}

void page_u8g2_settings_poweronconnect(){
    DrawConfig(draw_page_2, loop_u8g2_settings_poweronconnect);
}

void page_u8g2_poweroff(){
    DrawConfig(draw_page_2, loop_u8g2_poweroff);
}

void (*draw_map_x[])()={
    page_u8g2_mainmenu,         // 0
    page_u8g2_page1,            // 1
    page_u8g2_speed,            // 2
    page_u8g2_input,            // 3
    page_u8g2_connect,          // 4
    page_u8g2_network,          // 5
    page_u8g2_network_cont,     // 6
    page_u8g2_settings,         // 7
    page_u8g2_poweroff          // 8
};

void (*draw_map_y1[])()={ // page1, draw_map_y_sz=3
    page_u8g2_page1,
    page_u8g2_page2,
    page_u8g2_page3
};

void (*draw_map_y7[])()={ // settings, draw_map_y_sz=2
    page_u8g2_settings,
    page_u8g2_settings_poweronconnect
};

void check_and_process_horizontal_pages(unsigned dir){ // horizontal animation allocate.
    if (dir==JS_DIR_RIGHT){ // right
        if (++mapx_idx>=sizeof(draw_map_x)/sizeof(*draw_map_x)) mapx_idx=0;
        slide_dir=SLIDE_DIR_TIPRIGHT;
        slide_idx=0;
        draw_map_x[mapx_idx]();
        slide_2_pos=U8G2_X; // the order is of concern
    }
    else { // JS_DIR_LEFT, left
        if (--mapx_idx<0) mapx_idx=sizeof(draw_map_x)/sizeof(*draw_map_x)-1;
        slide_dir=SLIDE_DIR_TIPLEFT;
        slide_idx=0;
        draw_map_x[mapx_idx]();
        slide_2_pos=-U8G2_X; // the order is of concern
    }
}

bool check_and_process_vertical_pages(unsigned dir){ // vertical animation allocate.
    void (*(*the_v_pages))()=0;
    unsigned draw_map_y_sz=0;
    switch (mapx_idx){ // the pages which have vertical pages.
        case 1:
            the_v_pages=draw_map_y1;
            draw_map_y_sz=3;
            break;
        case 7:
            the_v_pages=draw_map_y7;
            draw_map_y_sz=2;
            break;
        default: return false;
    }
    if (dir==JS_DIR_UP){ // up
        if (--mapy_idx<0) mapy_idx=draw_map_y_sz-1;
        slide_dir=SLIDE_DIR_TIPUP;
        slide_idx=0;
        the_v_pages[mapy_idx]();
        slide_2_pos=-U8G2_Y; // the order is of concern
    }
    else { // JS_DIR_DOWN, down
        if (++mapy_idx>=draw_map_y_sz) mapy_idx=0;
        slide_dir=SLIDE_DIR_TIPDOWN;
        slide_idx=0;
        the_v_pages[mapy_idx]();
        slide_2_pos=U8G2_Y; // the order is of concern
    }
    return true;
}

void joystick_btn_functions(){ // joystick button response function.
    static unsigned to_upper=0; // character case. (0, lower); (1, upper once); (2, upper fixed).
    switch (mapx_idx){
        case 3: // input
            if (!locked_lr){
                locked_lr=1;
                jtk_pos_x=0;
                jtk_pos_y=0;
                //// to_upper=0;
                input_str="";
            }
            else if ((jtk_pos_x==5) && (jtk_pos_y==2)){ // save and exit
                locked_lr=0;
            }
            else if ((jtk_pos_x==4) && (jtk_pos_y==2)){ // backspace
                if (input_str.length()) input_str.remove(input_str.length()-1);
            }
            else if ((jtk_pos_x==3) && (jtk_pos_y==2)){ // clear all
                input_str="";
            }
            else if ((jtk_pos_x==2) && (jtk_pos_y==2)){ // to upper fixed
                if (to_upper!=2) to_upper=2;
                else to_upper=0;
            }
            else if ((jtk_pos_x==1) && (jtk_pos_y==2)){ // to upper once
                to_upper=1;
            }
            else if ((jtk_pos_x==0) && (jtk_pos_y==2)){ // line-feed
                input_str+="\r\n";
            }
            else {
                if (to_upper==1){String a(alphabet[jtk_pos_y][jtk_pos_x]); a.toUpperCase(); input_str+=a; to_upper=0;}
                else if (to_upper==2){String a(alphabet[jtk_pos_y][jtk_pos_x]); a.toUpperCase(); input_str+=a;}
                else input_str+=alphabet[jtk_pos_y][jtk_pos_x];
            }
            break;
        case 4: // connect
            copy_sta_do_the_connect^=1;
            break;
        case 7: // settings
            switch (mapy_idx){
                case 1: // page_u8g2_settings_poweronconnect
                    {
                        timer_draw_manip.Pause(); ///// pause the timer and the delay are all must else wdtreset.
                        delay(300); ///// pause the timer and the delay are all must else wdtreset.
                        unsigned a=0;
                        cDataBase s("/settings");
                        s.Open();
                        s.ReadWords(a, "copy_sta_do_the_connect", 0);
                        s.Close();
                        s.ReplaceWordsEx("copy_sta_do_the_connect", 0, a^=1);
                        timer_draw_manip.Resume();
                        DrawEnd(draw_page_1);
                        if (a) DrawConfig(draw_page_1, loop_u8g2_on);
                        else DrawConfig(draw_page_1, loop_u8g2_off);
                    }
                    break;
            }
            break;
        case 8: // poweroff
            DrawEnd(draw_page_1);
            DrawConfig(draw_page_1, loop_u8g2_powerdown);
            poweroff=1;
            break;
    }
}

void joystick_dir_funcs(unsigned by_the_map_idx, unsigned dir){ // joystick direction response function.
    switch (by_the_map_idx){
        case 3: // input
            if (jtk_pos_x>24) jtk_pos_x=24;
            else if (jtk_pos_x<0) jtk_pos_x=0;
            if (jtk_pos_y>2) jtk_pos_y=2;
            else if (jtk_pos_y<0) jtk_pos_y=0;
            DrawConfig(draw_page_1, loop_u8g2_alphabet_change);
            break;
    }
}

void DrawProcessing(){ // the main drawing function where according to joystick behaviors.

    // ------ first, need to get a useful of reproduced joystick behaviors, the joystick_dir.
    if (!joystick_action){
        joystick_dir=0;
        if (joystick_x_old!=joystick_tri[1]){joystick_x_old=joystick_tri[1];joystick_dir|=((joystick_x_old&1)? 1: 2);}
        if (joystick_y_old!=joystick_tri[0]){joystick_y_old=joystick_tri[0];joystick_dir|=((joystick_y_old&1)? 4: 8);}
        static unsigned breakcode=1;
        if (joystick_b_old!=joystick_tri[2]){joystick_b_old=joystick_tri[2]; if (breakcode){joystick_dir|=16; breakcode=0;}}
        else breakcode=1;
    }
    else {
        joystick_x_old=joystick_tri[1];
        joystick_y_old=joystick_tri[0];
        joystick_b_old=joystick_tri[2];
    }

    // ------ next, control the incoming event.
    if (slide_2_pos) return;
    if (draw_page_2){ // means done the sliding draw and to deallocate.
        DrawEnd(draw_page_1);
        draw_page_1=draw_page_2;
        draw_page_2=0;
        joystick_action=0;
        DrawSubmit=0; // change from 2 to 0, deallocate draw_page_2(1), in order to reenable the refresh
    }

    if (joystick_dir){
        unsigned i=1;
        do {
            if (slide_2_pos) break;
            switch (joystick_dir&i){
                case JS_DIR_IDLE: break;
                case JS_DIR_LEFT: // left
                    if (locked_lr){
                        jtk_pos_x--;
                        joystick_dir_funcs(mapx_idx, JS_DIR_LEFT);
                        joystick_action=0;
                        joystick_dir=0;
                        return;
                    }
                    mapy_idx=0; // y is submenu, so needs to reset.
                    check_and_process_horizontal_pages(JS_DIR_LEFT);
                    break;
                case JS_DIR_RIGHT: // right
                    if (locked_lr){
                        jtk_pos_x++;
                        joystick_dir_funcs(mapx_idx, JS_DIR_RIGHT);
                        joystick_action=0;
                        joystick_dir=0;
                        return;
                    }
                    mapy_idx=0; // y is submenu, so needs to reset.
                    check_and_process_horizontal_pages(JS_DIR_RIGHT);
                    break;
                case JS_DIR_DOWN: // down
                    if (locked_lr){
                        jtk_pos_y++;
                        joystick_dir_funcs(mapx_idx, JS_DIR_DOWN);
                        joystick_action=0;
                        joystick_dir=0;
                        return;
                    }
                    else if (!check_and_process_vertical_pages(JS_DIR_DOWN)){
                        joystick_action=0;
                        joystick_dir=0;
                        return;
                    }
                    break;
                case JS_DIR_UP: // up
                    if (locked_lr){
                        jtk_pos_y--;
                        joystick_dir_funcs(mapx_idx, JS_DIR_UP);
                        joystick_action=0;
                        joystick_dir=0;
                        return;
                    }
                    else if (!check_and_process_vertical_pages(JS_DIR_UP)){
                        joystick_action=0;
                        joystick_dir=0;
                        return;
                    }
                    break;
                case JS_DIR_PUSH: // push
                    joystick_btn_functions();
                    joystick_action=0;
                    joystick_dir=0;
                    return;
                default: break;
            }
        }
        while ((i<<=1)<=JS_DIR_PUSH); // 16

        joystick_action=1;
        joystick_dir=0;
    }
}


// ------------------------------------ http
void my_http_request(){ // called only if http_request_ip assigned.
    static unsigned tmo=micros();
    if (Timeout(433, tmo)){ // each 0.433 seconds to send http request.
        HTTPC_http_client_connect(web_link.c_str());
    }
}

void my_setup_http_server(){

    uri_list+=\
        "/version\r\n"\
        "/s (for ap mode only)\r\n"\
        "/stop (stop the timer)\r\n"\
        "/poweroff\r\n";

    server_p->on("/version", []() { // show version
        server_p->send(200, "text/plain", VERSION_202105180003);
    });

    server_p->on("/stop", []() { // stop the timer
        extern unsigned stop_the_timer;
        stop_the_timer=1;
        server_p->send(200, "text/plain", "Done");
    });

    server_p->on("/poweroff", []() {
        extern unsigned poweroff;
        poweroff=1;
        server_p->send(200, "text/plain", "Done");
        
    });

#if AP_MODE
    server_p->on("/s", []() {
        int i=(random(3)? -1: 1);
        static int j=0;
        static unsigned k=0;
        if (!k && random(17)>14) k=2;
        else if (k){k--; i=1;}
        j+=i;
        if (j<0) j=0;
        else if (j>7) j=7;
        ap_mode_speed_str=(j*1000+random(1000));
        server_p->send(200, "text/plain", ap_mode_speed_str.c_str());
    });
#endif
}


void init_params(){
    cDataBase s("/settings");
    s.Open();
    unsigned BootTimes=0;

    if (s.ReadWords(BootTimes, "BootTimes", 0)){
        printf("load params success: %s %u\r\n", "BootTimes", BootTimes);
    }
    s.ReadWords(copy_sta_do_the_connect, "copy_sta_do_the_connect", 0);

    s.Close();
    s.ReplaceWordsEx("BootTimes", 0, ++BootTimes);
    if (!(BootTimes%5)) s.Renew();
}


void setup(){
    // the order in setup() is of concern.
    delay(2000); // the delay 2 seconds is must by sil usb2ur chip.
    Serial.begin(115200);
    while (!Serial) delay(5);
    printf("\r\nwelcome\r\n\r\n\r\n");
    Serial.flush();

    setup_u8g2();
    DrawSetup();
    DrawConfig(draw_page_1, loop_u8g2_welcome);
    DrawSubmit=1;
    demandcall_draw_manip();

    ////pinMode(BUTTON_GPIO, INPUT_PULLUP); // not effective.
    pinMode(BUTTON_GPIO, OUTPUT);
    digitalWrite(BUTTON_GPIO, HIGH); // keep the power
    pinMode(MUX_S0_GPIO, OUTPUT); // mux gpio s0
    pinMode(MUX_S1_GPIO, OUTPUT); // mux gpio s1

    randomSeed(micros()%ESP.getChipId());
    sta_do_the_connect=0; // do not connect ap as default.
    copy_sta_do_the_connect=0; // do not connect ap as default. it is changed in init_params().
    init_params();
    setup_IoT();

    // ------ config the main menu
    if (!copy_sta_do_the_connect) delay(250);
    DrawEnd(draw_page_1);
    DrawConfig(draw_page_1, loop_u8g2_1);
    DrawConfig(draw_page_1, loop_u8g2_3);
    DrawConfig(draw_page_1, loop_u8g2_status_051);
    DrawConfig(draw_page_1, loop_u8g2_status_052);
    DrawConfig(draw_page_1, loop_u8g2_status_053);
    DrawConfig(draw_page_1, loop_u8g2_status_054);
    DrawConfig(draw_page_1, loop_u8g2_status_055);
    AutoDrawSubmit=1;
    timer_draw_manip.setTimer(BTN_PERIOD, isr_draw_manip_timer);
    timer_draw_manip.Resume();
}


IRAM_ATTR const char *bat_status="-"; // (-)=0, (f)ull=1, (d)ischarging=2, (c)harging=3, (l)ow=4.
void checkBattery(){ // better polling at 10 seconds
    static unsigned f_tmo=0;
    static unsigned battery_adc=joystick[3]; // the current battery ADC value
    if (*bat_status=='l') return; // 4
    if (battery_adc<joystick[3]){bat_status="c"; f_tmo=0;} // 3
    else if (battery_adc>joystick[3]){bat_status="d"; f_tmo=0;} // 2
    else {
        if (!f_tmo) f_tmo=micros();
        else if (Timeout(50000, f_tmo)){ // 50 seconds no change means full
            bat_status="f"; // 1
        }
    }
    battery_adc=joystick[3];
}


void loop(){

    // ------ iot processing. to reconnect is monopolized. so disable the timer first
    loop_IoT();
#if STA_MODE
    static unsigned iflag=0;
    if (!iflag && (is_station_mode_established!=copy_sta_do_the_connect)){
        timer_draw_manip.Pause();
        iflag=1;
        delay(300);
        sta_do_the_connect=copy_sta_do_the_connect;
    }
    else if (iflag && (is_station_mode_established==sta_do_the_connect)){
        timer_draw_manip.Resume();
        iflag=0;
        delay(300);
        if (is_station_mode_established){
            netstr_ssid=WiFi.SSID();
            netstr_remoteip=remote_ap_ip;
            netstr_localip=WiFi.localIP().toString();
            strcpy(http_request_ip, remote_ap_ip); // assume the ap is the server to http request.
            web_link=String("http://")+http_request_ip+"/s";

            #if defined(USE_MDNS_FULL)
                netstr_localhostname=mDnsHttpResponse();
            #elif defined(USE_MDNS_STANDALONE)
                netstr_localhostname=String("http://")+HOSTNAME+".local/";
            #endif // defined(USE_MDNS_FULL)
        }
    }
    static unsigned abort_conn=0;
    if (is_station_mode_established!=sta_do_the_connect){ // need to abort if always failed.
        static unsigned t=0;
        if (!abort_conn){
            abort_conn=30000; // 30 seconds
            t=micros();
        }
        else if (Timeout(abort_conn, t, true, true)){
            sta_do_the_connect^=1;
            copy_sta_do_the_connect=sta_do_the_connect;
            abort_conn=0;
        }
        return;
    }
    else abort_conn=0;
#endif // STA_MODE

    // ------ button sensing then draw processing
    static unsigned tmo4=micros();
    if (Timeout(200, tmo4)){
        DrawProcessing();
    }

    // ------ check the battery status at each 10 seconds
    static unsigned tmo3=micros();
    if (Timeout(10000, tmo3)) checkBattery();

    // ------ stop the timer in order to webupdate
    if (stop_the_timer){
        if (stop_the_timer==1){
            stop_the_timer=2;
            timer_draw_manip.Stop();
            pinMode(MUX_S1_GPIO, INPUT);
        }
    }

    // ------ after 60 seconds to power down if idle. note that whenever webupdate before, press a button to get free time.
    static unsigned tmo2=micros(), tmo22=tmo2;
    if (joystick_alive){
        joystick_alive=0;
        countdown=0;
        tmo2=micros();
    }
    else if (Timeout(50000, tmo2, true) && !countdown) countdown=9;
    else if (countdown){
        if (Timeout(1050, tmo22)) countdown--;
    }
    if ((countdown==1) || poweroff){
        timer_draw_manip.Pause();
        DrawEnd(draw_page_1);
        DrawConfig(draw_page_1, loop_u8g2_powerdown);
        demandcall_draw_manip();
        printf("\r\n[power down]\r\n");
        delay(1000);
        pinMode(MUX_S1_GPIO, INPUT);
        digitalWrite(BUTTON_GPIO, LOW);
        pinMode(BUTTON_GPIO, INPUT);
        system_deep_sleep(0);
        while (1);
    }

    // ------ debug
    /*static unsigned tmo1=micros();
    if (Timeout(1500, tmo1)){
        printf("%d %d %d %d %d[x[%d, %d], y[%d, %d], b[%d, %d], %d[%d, %fV]]\r\n", joystick_dir, slide_2_pos, joystick_action, jtk_pos_x, jtk_pos_y,\
        joystick[1], joystick_tri[1],\
        joystick[0], joystick_tri[0],\
        joystick[2], joystick_tri[2],\
        joystick[3], joystick_tri[3], BVOLTAGE(joystick[3]));
    }*/
}
